#include "ButtonMatrix.h"

ButtonMatrix::ButtonMatrix(byte *row, byte *col, byte rows, byte cols) {
  rowPins = row;
  columnPins = col;
  numRows = rows;
  numCols = cols;
  sizeBM.rows = numRows;
  sizeBM.columns = numCols;

  setDebounceTime(10);
  setHoldTime(500);
  buttonEventListener = 0;

  startTime = 0;
  single_button = false;
}

// Returns a single button only. Retained for backwards compatibility.
int ButtonMatrix::getButton() {
  single_button = true;

  if (getButtons() && button[0].stateChanged && (button[0].bstate == PRESSED))
    return button[0].bindex;

  single_button = false;

  return NO_BUTTON;
}

// Populate the button list.
bool ButtonMatrix::getButtons() {
  bool buttonActivity = false;

  // Limit how often the button matrix is scanned. This makes the loop() run 10 times as fast.
  if ( (millis() - startTime) > debounceTime ) {
    scanButtons();
    buttonActivity = updateList();
    startTime = millis();
  }

  return buttonActivity;
}

// Private : Hardware scan
void ButtonMatrix::scanButtons() {
  // Re-intialize the row pins. Allows sharing these pins with other hardware.
  for (byte r = 0; r < sizeBM.rows; r++) {
    pin_mode(rowPins[r], INPUT_PULLUP);
  }

  // bitMap stores ALL the buttons that are being pressed.
  for (byte c = 0; c < sizeBM.columns; c++) {
    pin_mode(columnPins[c], OUTPUT);
    pin_write(columnPins[c], LOW);	// Begin column pulse output.
    for (byte r = 0; r < sizeBM.rows; r++) {
      bitWrite(bitMap[r], c, !pin_read(rowPins[r]));  // button press is active low so invert to high.
    }
    // Set pin to high impedance input. Effectively ends column pulse.
    pin_write(columnPins[c], HIGH);
    pin_mode(columnPins[c], INPUT);
  }
}

// Manage the list without rearranging the buttons. Returns true if any buttons on the list changed state.
bool ButtonMatrix::updateList() {

  bool anyActivity = false;

  // Delete any IDLE buttons
  for (byte i = 0; i < numButtons(); i++) {
    if (button[i].bstate == IDLE) {
      button[i].bindex = NO_BUTTON;
      button[i].stateChanged = false;
    }
  }

  // Add new buttons to empty slots in the button list.
  for (byte r = 0; r < sizeBM.rows; r++) {
    for (byte c = 0; c < sizeBM.columns; c++) {
      boolean bbutton = bitRead(bitMap[r], c);
      int buttonIndex = c * sizeBM.rows + r; //r * sizeBM.columns + c;
      int idx = findInList (buttonIndex);
      // button is already on the list so set its next state.
      if (idx > -1)	{
        nextButtonState(idx, bbutton);
      }
      // Button is NOT on the list so add it.
      if ((idx == -1) && bbutton) {
        for (byte i = 0; i < numButtons(); i++) {
          if (button[i].bindex == NO_BUTTON) {		// Find an empty slot or don't add button to list.
            button[i].bindex = buttonIndex;
            button[i].bstate = IDLE;		// Buttons NOT on the list have an initial state of IDLE.
            nextButtonState (i, bbutton);
            break;	// Don't fill all the empty slots with the same button.
          }
        }
      }
    }
  }

  // Report if the user changed the state of any button.
  for (byte i = 0; i < numButtons(); i++) {
    if (button[i].stateChanged) anyActivity = true;
  }

  return anyActivity;
}

// Private
// This function is a state machine but is also used for debouncing the buttons.
void ButtonMatrix::nextButtonState(byte idx, boolean bbutton) {
  button[idx].stateChanged = false;

  switch (button[idx].bstate) {
    case IDLE:
      if (bbutton == CLOSED) {
        transitionTo (idx, PRESSED);
        holdTimer = millis();
      }		// Get ready for next HOLD state.
      break;
    case PRESSED:
      if ((millis() - holdTimer) > holdTime)	// Waiting for a button HOLD...
        transitionTo (idx, HOLD);
      else if (bbutton == OPEN)				// or for a button to be RELEASED.
        transitionTo (idx, RELEASED);
      break;
    case HOLD:
      if (bbutton == OPEN)
        transitionTo (idx, RELEASED);
      break;
    case RELEASED:
      transitionTo (idx, IDLE);
      break;
  }
}

bool ButtonMatrix::isPressed(int buttonIndex) {
  for (byte i = 0; i < numButtons(); i++) {
    if ( button[i].bindex == buttonIndex ) {
      if ( (button[i].bstate == PRESSED) && button[i].stateChanged )
        return true;
    }
  }
  return false;	// Not pressed.
}

// Search by index for a button in the list of active buttons.
// Returns -1 if not found, returns the active buttons list index if found.
int ButtonMatrix::findInList (int buttonIndex) {
  for (byte i = 0; i < numButtons(); i++) {
    if (button[i].bindex == buttonIndex) {
      return i;
    }
  }
  return -1;
}

int ButtonMatrix::waitForButton() {
  int waitButton = NO_BUTTON;
  while ( (waitButton = getButton()) == NO_BUTTON );	// Block everything while waiting for a buttonpress.
  return waitButton;
}

ButtonState ButtonMatrix::getState() {
  return button[0].bstate;
}

// The end user can test for any changes in state before deciding
// if any variables, etc. needs to be updated in their code.
bool ButtonMatrix::buttonStateChanged() {
  return button[0].stateChanged;
}

// The number of buttons on the button list, equals the number
// of bytes in the button list divided by the number of bytes in a Button object.
byte ButtonMatrix::numButtons() {
  return numRows * numCols;
  //return sizeof(button)/sizeof(Button);
}

// Minimum debounceTime is 1 ms. Any lower *will* slow down the loop().
void ButtonMatrix::setDebounceTime(uint debounce) {
  debounce < 1 ? debounceTime = 1 : debounceTime = debounce;
}

void ButtonMatrix::setHoldTime(uint hold) {
  holdTime = hold;
}

void ButtonMatrix::addEventListener(void (*listener)(int)) {
  buttonEventListener = listener;
}

void ButtonMatrix::transitionTo(byte idx, ButtonState nextState) {
  button[idx].bstate = nextState;
  button[idx].stateChanged = true;

  // Sketch used the getButton() function.
  // Calls buttonEventListener only when the first button in slot 0 changes state.
  if (single_button)  {
    if ( (buttonEventListener != NULL) && (idx == 0) )  {
      buttonEventListener(0);
    }
  }
  // Sketch used the getButtons() function.
  // Calls buttonEventListener on any button that changes state.
  else {
    if (buttonEventListener != NULL)  {
      buttonEventListener(idx);
    }
  }
}
