#include "ButtonMatrix.h"
#include "LinkedList.h"
#include <math.h>

const int LED_PIN = 6;                                                      // onboard indicator LED (below "PA" pins on Teensy++ 2.0)
const byte NUM_AXIS = 5;                                                    // five axis
const float SMOOTH_FACTOR = 10.0;                                           // Smoothing Factor for Axis Readout
const int MIN_AXIS_DELTA = 4;                                               // Minimum difference between old and new Axis Value, in order to avoid "fluttering" values
const byte ROWS = 10;                                                       // four rows
const byte COLS = 8;                                                        // three columns
const long SWITCHBUTTON_HOLD_TIME = 200;                                    // 200ms until a "Switch Button" will be released again
const long FORCE_UPDATE_TIME = 1000;
byte rowPins[ROWS] = {8, 9, 10, 11, 12, 13, 14, 15, 16, 17};                // connect to the row pinouts of the keypad
byte colPins[COLS] = {27, 0, 1, 2, 3, 4, 5, 7};                             // connect to the column pinouts of the keypad
int oldAxis[NUM_AXIS];                                                      // Old Delta-Comparison Axis Values
float fAxis[NUM_AXIS];                                                      // Axis Readout Smoothing Reference Value

int16_t buttonDirs[8] = {0, 45, 90, 135, 180, 225, 270, 315};               // Directions for POV Hat
ButtonMatrix buttonmatrix = ButtonMatrix( rowPins, colPins, ROWS, COLS );   // Our Button Matrix, 8 Rows x 10 Columns

unsigned long loopCount;                                                    // loops elapsed in total
unsigned long startTime;                                                    // Milliseconds of sketch starting time
unsigned long ledTime;                                                      // Milliseconds of last onboard LED activation
unsigned long joystickTime;                                                 // Milliseconds of last Joystick Axis Update
unsigned long forceUpdateTime;                                              // Milliseconds of last Forced Regular Update
bool ledOn;                                                                 // Current onboard LED Status
bool updateRequired;                                                        // Update of 1st Joystick Device USB Telegram pending
bool update2Required;                                                       // Update of 2nd Joystick Device USB Telegram pending
String msg;                                                                 // Serial Debugging Message
int analogTemp;                                                             // Temporary Value holding analog Axis Readout

class SwitchButton {
  public:
    int buttonIndex;
    unsigned long pressedTime;
};

LinkedList<SwitchButton*> mySwitchButtonList = LinkedList<SwitchButton*>();

void SwitchButtonPressed(int buttonIndex) {
  SwitchButton *theButton = new SwitchButton();
  theButton->buttonIndex = buttonIndex;
  theButton->pressedTime = millis();
  mySwitchButtonList.add(theButton);
}

int CheckPressedSwitchButtons() {
  if (mySwitchButtonList.size() < 1) return 0;
  int retVal = 0;
  while (true) {
    SwitchButton *theButton = mySwitchButtonList.get(0);
    if (millis() < theButton->pressedTime + SWITCHBUTTON_HOLD_TIME) break;
    if (theButton->buttonIndex < 32) { // 1st Joystick Device Buttons are index 0-31
      Joystick.button(theButton->buttonIndex + 1, 0); // Button is released
      retVal = 1;
    } else if (theButton->buttonIndex >= 40 && theButton->buttonIndex < 72) { // 2nd Joystick Device Buttons are index 40-71
      Joystick2.button(theButton->buttonIndex + 1, 0); // Button is released
      retVal = 2;
    }
    delete(theButton);
    mySwitchButtonList.remove(0);
    if (mySwitchButtonList.size() < 1) break;
  }
  return retVal;
}

void setup() {
  // Make sure that we have full control over Joystick USB telegrams
  Joystick.useManualSend(true);
  Joystick2.useManualSend(true);

  pinMode(LED_PIN, OUTPUT); // Enable onboard LED pin

  // Init serial Debugging
  Serial.begin(9600); // Always use 9600 baud
  Serial.println(" ******************");
  Serial.println("*                  *");
  Serial.println("* Mike's Switchbox *");
  Serial.println("*                  *");
  Serial.println(" ******************");
  Serial.println("");
  Serial.print("Number of Buttons: ");
  Serial.println(buttonmatrix.numButtons());

  // Init global variables
  loopCount = 0; // loops elapsed in total
  startTime = millis(); // Milliseconds of sketch starting time
  ledTime = millis(); // Milliseconds of last onboard LED activation
  joystickTime = millis(); // Milliseconds of last Joystick Axis Update
  msg = ""; // Serial Debugging Message
  ledOn = false; // Current onboard LED Status

  // Init old (delta comparison) and smoothing Axis values. 512 is Axis center
  for (int i = 0; i < NUM_AXIS; i++) {
    oldAxis[i] = 512;
    fAxis[i] = 512.0;
  }

  analogTemp = 0; // Temporary Value holding analog Axis Readout
  updateRequired = false; // Update of 1st Joystick Device USB Telegram pending
  update2Required = false; // Update of 2nd Joystick Device USB Telegram pending

  // Init Buttons
  buttonmatrix.getButtons(); // Initial Button Readout. Some Buttons might already be pressed at startup.
  for (int i = 0; i < buttonmatrix.numButtons(); i++) {
    if (buttonmatrix.button[i].bstate == PRESSED || buttonmatrix.button[i].bstate == HOLD) {
      if (i<20) SwitchButtonPressed(i);
      if (i < 32) { // 1st Joystick Device Buttons are index 0-31
        Joystick.button(i + 1, 1); // Button is pressed
        //SwitchButtonPressed(i);
      } else if (i >= 40 && i < 72) { // 2nd Joystick Device Buttons are index 40-71
        Joystick2.button(i + 1, 1); // Button is pressed
        //SwitchButtonPressed(i);
      }
    } else {
      if (i < 32) // 1st Joystick Device Buttons are index 0-31
        Joystick.button(i + 1, 0); // Button is released
      else if (i >= 40 && i < 72) // 2nd Joystick Device Buttons are index 40-71
        Joystick2.button(i + 1, 0); // Button is released
    }
  }

  // Init Axis on Joystick Device No.1, 512 is Center
  Joystick.X(512);
  Joystick.Y(512);
  Joystick.Z(512);
  Joystick.Xrotate(512);
  Joystick.Yrotate(512);
  // Init Axis on Joystick Device No.2, 512 is Center
  Joystick2.X(512);
  Joystick2.Y(512);
  Joystick2.Z(512);
  Joystick2.Xrotate(512);
  Joystick2.Yrotate(512);

  // POV Hat Init is actually simplified, we don't care whether any of the regarding buttons
  // is actually pressed at startup or not.
  // This means that POV Hats should not be used for toggle switches.
  // They will work fine for momentare switches and Pushbuttons though.
  // Init POV Hat on Joystick Device No.1, -1 is Center
  Joystick.hat(1, -1);
  Joystick.hat(2, -1);
  Joystick.hat(3, -1);
  // Init POV Hat on Joystick Device No.2, -1 is Center
  Joystick2.hat(1, -1);
  Joystick2.hat(2, -1);
  Joystick2.hat(3, -1);

  // Send initializing Joystick Telegrams for both devices
  Joystick.send_now();
  Joystick2.send_now();

  buttonmatrix.addEventListener(buttonEvent); // Add an event listener for this button matrix
}

void loop() {

  // Log avarage loops per second on Debugging interface, every 30 seconds
  loopCount++;
  if ( (millis() - startTime) > 30000 ) {
    Serial.print("Average loops per second = ");
    Serial.println(loopCount / 30);
    startTime = millis();
    loopCount = 0;
  }

  // Any button/switch activity will briefly light up the onboard LED for reference
  if (ledOn && (millis() - ledTime) > 1 ) {
    ledOn = false;
    ledTime = millis();
    digitalWrite(LED_PIN, LOW);
  }

  // Check for any button/switch activity. If there's any status change, it will be dealt with in the Event Listener "buttonEvent"
  buttonmatrix.getButtons();

  if ( (millis() - joystickTime) > 10 ) { // Every 10 Milliseconds, it's time to update the Joystick Axis and send new USB Telegrams if any are pending
    joystickTime = millis();

    bool forceUpdate = false;
    if (millis() - forceUpdateTime > FORCE_UPDATE_TIME) {
      forceUpdate = true;
      forceUpdateTime = millis();
    }

    for (int i = 0; i < NUM_AXIS; i++) { // Check all Axis
      fAxis[i] = (fAxis[i] * (SMOOTH_FACTOR - 1.0) + (float)analogRead(i)) / SMOOTH_FACTOR; // Smoothen the Axis Readout
//      analogTemp = analogRead(i); // Store Axis value in temp variable
      bool isOnEdge = false;
      if (fAxis[i] < 10.0) {
        fAxis[i] = 0.0;
        if ((oldAxis[i] < (MIN_AXIS_DELTA * 2)) && (oldAxis[i] > 0)) isOnEdge = true;
      }
      if (fAxis[i] > 1013.0) {
        fAxis[i] = 1023.0;
        if ((oldAxis[i] > (1023 - (MIN_AXIS_DELTA * 2))) && (oldAxis[i] < 1023)) isOnEdge = true;
      }
      analogTemp = round(fAxis[i]); // Store Axis integer value in temp variable
      
      if (isOnEdge || forceUpdate || abs(analogTemp - oldAxis[i]) >= MIN_AXIS_DELTA) { // Check if raw delta is smaller then min delta
        if (!isOnEdge && abs(analogTemp - oldAxis[i]) < MIN_AXIS_DELTA) continue; // Below Delta now? Don't care...

        oldAxis[i] = analogTemp; // New Reference Value for Delta Comparison on this Axis
        switch (i) {
          case 0:
            Joystick.X(analogTemp); // Update Both Joystick's Axis
            Joystick2.X(analogTemp); // to the new value (not necessary, just for the sake of simplicity)
            updateRequired = true; // We need to update Joystick Device No. 1
            update2Required = true; // and 2 to reflect Axis Changes
            break;
          case 1:
            Joystick.Y(analogTemp);
            Joystick2.Y(analogTemp);
            updateRequired = true;
            update2Required = true;
            break;
          case 2:
            Joystick.Z(analogTemp);
            Joystick2.Z(analogTemp);
            updateRequired = true;
            update2Required = true;
            break;
          case 3:
            Joystick.Xrotate(analogTemp);
            Joystick2.Xrotate(analogTemp);
            updateRequired = true;
            update2Required = true;
            break;
          case 4:
            Joystick.Yrotate(analogTemp);
            Joystick2.Yrotate(analogTemp);
            updateRequired = true;
            update2Required = true;
            break;
          default: // Funny thing... this Axis doesn't even exist!
            break;
        }
      }
    }

    switch (CheckPressedSwitchButtons()) {
      case 1:
        updateRequired = true;
        break;
      case 2:
        update2Required = true;
        break;
      default:
       break;
    }

    if (updateRequired) { // New Telegram for Joystick Device No.1
      updateRequired = false;
      Joystick.send_now();
    }
    if (update2Required) { // New Telegram for Joystick Device No.2
      update2Required = false;
      Joystick2.send_now();
    }

  }
}

// Button Matrix Even Handler.
// If any Button or Switch State changes, this Event Handler will get called.
// "listIndex" is the Index of an internal List of the ButtonMatrix holding buttons/switches with
// changed state.
void buttonEvent(ButtonEvent listIndex) {
  switch (buttonmatrix.button[listIndex].bstate) { // Check new State
    case PRESSED: // Button has been pressed (or Switch is in active State)
      msg = " PRESSED.";
      break;

    case RELEASED: // Button has been released (or Switch is in inactive State)
      msg = " RELEASED.";
      break;

    case HOLD: // Dont care for continuous Button Presses or Idling, or anything unknown.
    case IDLE:
    default:
      return;
  }

  // Turn on the Onboard LED because something happened on the Buttons/Switches!
  ledOn = true;
  ledTime = millis();
  digitalWrite(LED_PIN, HIGH);

  // Output new State on serial Debugging line
  Serial.print("Button ");
  Serial.print(buttonmatrix.button[listIndex].bindex);
  Serial.println(msg);

  int buttonIndex = buttonmatrix.button[listIndex].bindex; // Get the index of our 80-button array
  bool buttonPressed = false;
  if (buttonmatrix.button[listIndex].bstate == PRESSED || buttonmatrix.button[listIndex].bstate == HOLD) { // Button is pressed
    buttonPressed = true;
  }

  if (buttonIndex < 20) SwitchButtonPressed(buttonIndex);
  if (buttonIndex < 32) { // These are directly mapped buttons of Device No. 1
    if (buttonPressed) {
      Joystick.button(buttonIndex + 1, 1);
      //SwitchButtonPressed(buttonIndex);
    } else {
      Joystick.button(buttonIndex + 1, 0);
    }
    updateRequired = true; // Device No.1 needs a USB telegram Update.
  }
  else if (buttonIndex < 40) { // These are buttons of Device No. 1 mapped to POV Hat No. 3 on that device
    if (buttonPressed) {
      Joystick.hat(2, buttonDirs[buttonIndex - 32]); // Hat Index 2 is the 2nd-last Hat on the Device. As we have 4 hats, this is the third.
    } else {
      Joystick.hat(2, -1); // Device No.1 needs a USB telegram Update.
    }
    updateRequired = true;
  }
  if (buttonIndex < 72) { // These are directly mapped buttons of Device No. 2
    if (buttonPressed) {
      Joystick2.button(buttonIndex - 39, 1);
      //SwitchButtonPressed(buttonIndex);
    } else {
      Joystick2.button(buttonIndex - 39, 0);
    }
    update2Required = true; // Device No.2 needs a USB telegram Update.
  }
  else if (buttonIndex < 80) { // These are buttons of Device No. 2 mapped to POV Hat No. 3 on that device
    if (buttonPressed) {
      Joystick2.hat(1, buttonDirs[buttonIndex - 72]); // Hat Index 1 is the last Hat on the Device. As we have 4 hats, this is the fourth.
    } else {
      Joystick2.hat(1, -1);
    }
    update2Required = true; // Device No.2 needs a USB telegram Update.
  }
}
